/**
Interface für Datenelemente

- author: Klaus Reinold
- version: 1.0
*/
protocol Datenelement: AnyObject
{
    /** Methode zur Ausgabe von Information über das Datenelement */
    func InformationAusgeben()
    
    /**
    Gibt zurück, ob der Name gleich den übergebenen Wert ist.
    - parameters:
        - gesuchterWert: gesuchter Wert
    - returns: Wahrheitswert, der angibt, ob der Name der gesuchte ist.
    */
    func SchlüsselIstGleich(gesuchterWert: String) -> Bool
        
    /**
    Vergleich zweier Datenelemente bezüglich einer Ordnungsrelation
    - parameters:
        - dvergleich: Datenelement zum Vergleich
    */
    func IstKleinerAls(dvergleich: Datenelement) -> Bool
}

/**
Klasse zur Verwaltung einer Lottozahl
 
- author: Klaus Reinold
- version: 1.0
*/
class Zahl: Datenelement
{
    /**Zu verwaltender Wert*/
    private var z: Int

    /**
    Konstruktor
    - parameters:
        - zahlNeu:Wert der Zahl
    */
    init(zahlNeu: Int)
    {
        z = zahlNeu
    }

    /**
    Ausgabe der Zahl auf der Konsole
    */
    func InformationAusgeben()
    {
        print(z)
    }

    /**
    Ausgabe der verwalteten Zahl
    - returns: Zahl
    */
    func ZahlGeben() -> Int
    {
        return z
    }
    
    /**
    Gibt zurück, ob der Name gleich den übergebenen Wert ist.
    - parameters:
        - gesuchterWert: gesuchter Wert
    - returns: Wahrheitswert, der angibt, ob der Name der gesuchte ist.
    */
    func SchlüsselIstGleich(gesuchterWert: String) -> Bool
    {
        return "\(z)" == gesuchterWert
    }
    
    /**
    Vergleich zweier Datenelemente bezüglich einer Ordnungsrelation
    - parameters:
        - dvergleich: Datenelement zum Vergleich
    */
    func IstKleinerAls(dvergleich: Datenelement)
    {
        let zahl = dvergleich as! Zahl
        if z < zahl.ZahlGeben()
        {
            return true
        }
        else
        {
            return false
        }
    }
}


/**
 Abstrakte Klasse Listenelement als Oberklasse für Knoten und Abschluss

- author: Klaus Reinold
- version: 1.0
*/
class Listenelement
{

    /**
    Bestimmt die Länge der restlichen Liste ab dem aktuellen Knoten
    - returns: Länge der restlichen Liste
     ´*/
    func RestlängeGeben() -> Int
    {
        fatalError("Listenelement.RestlängeGeben muss überschrieben werden")
    }
    
    /**
     Gibt das Datenelement des Vorgängers (also des letzten Knotens) zurück
    - parameters:
        - dNeu übergebenes Datenelement
    - returns: hinterstes Datenelement
    */
    func EndeGeben(dNeu: Datenelement?) -> Datenelement?

    /**
    Entfernt einen Knoten, der über sein Datenelement identifiziert wird, aus der Liste
    - parameters:
        - dvergleich: gesuchtes Datenelement
    - returns: Weitergabe des künftigen Nachfolgers
    */
    func KnotenEntfernen(dvergleich: Datenelement?) -> Listenelement
    {
        fatalError("Listenelement.KnotenEntfernen muss überschrieben werden")
    }
    
    /**
    Gibt das Datenelement eines Knotens zurück
    - returns: Datenelement
    */
    func DatenelementGeben() -> Datenelement?
    {
        fatalError("Listenelement.DatenelementGeben muss überschrieben werden")
    }

    /**
    Gibt den Nachfolger des Knoten aus.
    - returns: Nachfolger
    */
    func NachfolgerGeben() -> Listenelement
    {
        fatalError("Listenelement.NachfolgerGeben muss überschrieben werden")
    }
    
    /**
     Gibt die Information zu den Datenelementen auf der Konsole aus
    */
    func InformationAusgeben()
    {
        fatalError("Listenelement.InformationAusgeben muss überschrieben werden")
    }

    /**
    Fügt ein Datenelement am Ende der Liste ein
    - parameters:
        - dNeu: neues Datenelement
    */
    func HintenEinfügen(dNeu: Datenelement) -> Knoten
    {
        fatalError("Listenelement.HintenEinfügen muss überschrieben werden")
    }

    /**
    Fügt ein Datenelement vor einem bestimmten anderen Datenelement ein
    - parameters:
        - dNeu: neues Datenelement
        - dVergleich: gesuchtes Datenelement, vor dem eingefügt werden soll
    - returns: Weitergabe des künftigen Nachfolgers
    */
    func EinfügenVor(dNeu: Datenelement, dVergleich: Datenelement) -> Knoten
    {
        fatalError("Listenelement.EinfügenVor muss überschrieben werden")
    }
    
    /**
    Sucht ein Datenelement mit einem bestimmten Schlüsselwert in der Liste
    - parameters:
        - schlüssel: Zeichenkette, die als Schlüssel gefunden werden soll.
    - returns: das gesuchte Datenelement
    */
    func Suchen(schlüssel: String) -> Datenelement?
    {
        fatalError("Listenelement.Suchen muss überschrieben werden")
    }

    /**
     Sortiertes Einfügen in die Liste
    - parameters:
        - dNeu: neues Datenelement
    - returns: Knoten als neuer Nachfolger für den Vorgänger
    */
    func SortiertEinfügen(dNeu: Datenelement) -> Knoten
    {
        fatalError("Listenelement.SortiertEinfügen muss überschrieben werden")
    }
}

/**
 Klasse zur Kennzeichnung des Endes der Liste

- author: Klaus Reinold
- version: 1.0
*/
class Abschluss: Listenelement
{

    /**
    Bestimmt die Länge der restlichen Liste ab dem aktuellen Knoten
    - returns: Länge der restlichen Liste
    */
    override func RestlängeGeben() -> Int
    {
        return 0
    }
    
    /**
    Gibt das Datenelement des letzten Knotens aus
     - parameters:
         - dNeu übergebenes Datenelement
    - returns: übergebenes Datenelement
    */
    override func EndeGeben(dNeu: Datenelement?) -> Datenelement?
    {
        return dNeu
    }

    /**
    Entfernt einen Knoten, der über sein Datenelement identifiziert wird, aus der Liste
    - parameters:
        - dvergleich: gesuchtes Datenelement
    - returns: Weitergabe des künftigen Nachfolgers
    */
    override func KnotenEntfernen(dvergleich: Datenelement?) -> Listenelement
    {
        return self
    }
    
    /**
    Gibt das Datenelement eines Knotens zurück
    - returns: Datenelement
    */
    override func DatenelementGeben() -> Datenelement?
    {
        return nil
    }

    /**
     Da der Abschluss keinen Nachfolger hat, gibt er sich selbst zurück
    - returns: Nachfolger
    */
    override func NachfolgerGeben() -> Listenelement
    {
        return self
    }
    
    /**
     Gibt die Information zu den Datenelementen auf der Konsole aus
    */
    override func InformationAusgeben()
    {
    }

    /**
    Fügt ein Datenelement am Ende der Liste ein
    - parameters:
        - dNeu: neues Datenelement
    */
    override func HintenEinfügen(dNeu: Datenelement) -> Knoten
    {
        return Knoten(dNeu: dNeu, lNeu: self)
    }

    /**
    Fügt ein Datenelement vor einem bestimmten anderen Datenelement ein
    - parameters:
        - dNeu: neues Datenelement
        - dVergleich: gesuchtes Datenelement, vor dem eingefügt werden soll
    - returns: Weitergabe des künftigen Nachfolgers
    */
    override func EinfügenVor(dNeu: Datenelement, dVergleich: Datenelement) -> Knoten
    {
        return Knoten(dNeu: dNeu, lNeu: self)
    }
    
    /**
    Sucht ein Datenelement mit einem bestimmten Schlüsselwert in der Liste
    - parameters:
        - schlüssel: Zeichenkette, die als Schlüssel gefunden werden soll.
    - returns: das gesuchte Datenelement
    */
    override func Suchen(schlüssel: String) -> Datenelement?
    {
        return nil
    }

    /**
     Sortiertes Einfügen in die Liste
    - parameters:
        - dNeu: neues Datenelement
    - returns: Knoten als neuer Nachfolger für den Vorgänger
    */
    override func SortiertEinfügen(dNeu: Datenelement) -> Knoten
    {
        return Knoten(dNeu: dNeu, lNeu: self)
    }
}

/**
Beschreibung der Klasse Knoten

- author: Klaus Reinold
- version: 1.0
*/
class Knoten: Listenelement
{
    /** Verwaltung des nachfolgenden Knoten in der Liste. */
    private var nachfolger: Listenelement
    /** Verwaltung der Daten des Patienten */
    private var daten: Datenelement

    /**
    Konstruktor des Knoten
    - parameters:
        - dNeu: Datenelement
        - lNeu: neuer Nachfolger
    */
    init(dNeu: Datenelement, lNeu: Listenelement)
    {
        nachfolger  = lNeu
        daten = dNeu
    }

    /**
    Bestimmt die Länge der restlichen Liste ab dem aktuellen Knoten
    - returns: Länge der restlichen Liste
     ´*/
    override func RestlängeGeben() -> Int
    {
        return nachfolger.RestlängeGeben() + 1
    }
    
    /**
    Gibt das Datenelement des letzten Knotens aus
    - parameters:
        - dNeu: übergebenes Datenelement
    - returns: hinterstes Datenelement
    */
    override func EndeGeben(dNeu: Datenelement?) -> Datenelement?
    {
        return nachfolger.EndeGeben(dNeu: daten)
    }

    /**
     Entfernt den Knoten, dessen Datenelement das übergebene ist.
    - parameters:
        - dvergleich: gesuchtes Datenelement
    - returns: Weitergabe des künftigen Nachfolgers
    */
    override func KnotenEntfernen(dvergleich: Datenelement?) -> Listenelement
    {
        if daten === dvergleich
        {
            return nachfolger
        }
        else
        {
            nachfolger=nachfolger.KnotenEntfernen(dvergleich: dvergleich)
            return self
        }
    }

    /**
    Gibt den Nachfolger des Knoten aus.
    - returns: Nachfolger
    */
    override func NachfolgerGeben() -> Listenelement
    {
        return nachfolger
    }
    
    /**
    Gibt das Datenelement zurück
    - returns: Datenelement
    */
    override func DatenelementGeben() -> Datenelement
    {
        return daten
    }
    
    /**
     Gibt die Information zu den Datenelementen auf der Konsole aus
    */
    override func InformationAusgeben()
    {
        daten.InformationAusgeben()
        nachfolger.InformationAusgeben()
    }

    /**
    Fügt ein Datenelement am Ende der Liste ein
    - parameters:
        - dNeu: neues Datenelement
    */
    override func HintenEinfügen(dNeu: Datenelement) -> Knoten
    {
        nachfolger = nachfolger.HintenEinfügen(dNeu: dNeu)
        return self
    }

    /**
    Fügt ein Datenelement vor einem bestimmten anderen Datenelement ein
    - parameters:
        - dNeu: neues Datenelement
        - dVergleich: gesuchtes Datenelement, vor dem eingefügt werden soll
    - returns: Weitergabe des künftigen Nachfolgers
    */
    override func EinfügenVor(dNeu: Datenelement, dVergleich: Datenelement) -> Knoten
    {
        if daten === dVergleich
        {
            return Knoten(dNeu: dNeu, lNeu: self)
        }
        else
        {
            nachfolger = nachfolger.EinfügenVor(dNeu: dNeu, dVergleich: dVergleich)
            return self
        }
    }
    
    /**
    Sucht ein Datenelement mit einem bestimmten Schlüssel in der Liste
    - parameters:
        - schlüssel: Zeichenkette, die als Schlüssel gefunden werden soll.
    - returns: das gesuchte Datenelement
    */
    override func Suchen(schlüssel: String) -> Datenelement?
    {
        if daten.SchlüsselIstGleich(gesuchterWert: schlüssel)
        {
            return daten
        }
        else
        {
            return nachfolger.Suchen(schlüssel: schlüssel)
        }
    }

    /**
    Fügt ein Datenelement sortiert in die Liste ein
    - parameters:
        - dNeu: neues Datenelement
    - returns: Weitergabe des künftigen Nachfolgers
    */
    override func SortiertEinfügen(dNeu: Datenelement) -> Knoten
    {
        if daten.IstKleinerAls(dvergleich: dNeu)
        {
            nachfolger = nachfolger.SortiertEinfügen(dNeu: dNeu)
            return self
        }
        else
        {
            return Knoten(dNeu: dNeu, lNeu: self)
        }
    }
}

/**
Beschreibung der Liste

- author: Klaus Reinold
- version: 1.0
*/
class Liste
{
    /** erstes Element der Liste */
    private var anfang: Listenelement

    /**
    Konstruktor der Liste
    */
    init()
    {
        anfang = Abschluss
    }

    /**
    Fügt ein neues Datenelement am Anfang der Liste ein.
    - parameters:
        - dNeu: neues Datenelement
    */
    func VorneEinfügen(dNeu: Datenelement)
    {
        anfang = Knoten(dNeu: dNeu, lNeu: anfang)
    }

    /**
    Bestimmt die Länge der Liste
    - returns: Listenlänge
    */
    func LängeGeben() -> Int
    {
        return anfang.RestlängeGeben()
    }
    
    /**
    Entfernt den letzten Knoten und gibt dessen Datenelement zurück.
    - returns: Datenelement des bisher letzten Knotens
    */
    func EndeEntfernen() -> Datenelement?
    {
        let dletzter = anfang.EndeGeben(dNeu: nil)
        anfang = anfang.KnotenEntfernen(dvergleich: dletzter)
        return dletzter
    }

    /**
    Entfernt den ersten Knoten aus der Liste und gibt sein Datenelement aus.
    - returns: Datenelement des bisher ersten Knotens
    */
    func AnfangEntfernen() -> Datenelement?
    {
        let d = anfang.DatenelementGeben()
        anfang = anfang.NachfolgerGeben()
        return d
    }

    /**
    Gibt Information über die Liste aus.
    */
    func InformationAusgeben()
    {
        anfang.InformationAusgeben()
    }

    /**
     Fügt einen neuen Patienten am Ende der Liste ein.
    - parameters:
        - dNeu: neues Datenelement
    */
    func HintenEinfügen(dNeu: Datenelement)
    {
        anfang = anfang.HintenEinfügen(dNeu: dNeu)
    

    /**
    Fügt ein neues Datenelement vor einem bestimmten Datenelement ein.
    - parameters:
        - dNeu: neues Datenelement
        - dVergleich: Datenelement, vor dem eingefügt werden soll
    */
    func EinfügenVor(dNeu: Datenelement, dVergleich: Datenelement)
    {
        anfang = anfang.EinfügenVor(dNeu: dNeu, dVergleich: dVergleich)
    }

    /**
    Sucht ein Datenelement mit einem bestimmten Schlüssel in der Liste
    - parameters:
        - schlüssel: Zeichenkette, die als Schlüssel gefunden werden soll.
    - returns: das gesuchte Datenelement
    */
    func Suchen(schlüssel: String) -> Datenelement?
    {
        return anfang.Suchen(schlüssel: schlüssel)
    }

    /**
    Fügt ein neues Datenelement sortiert in die Liste ein.
    - parameters:
        - dNeu: neues Datenelement
    */
    func SortiertEinfügen(dNeu: Datenelement)
    {
        anfang = anfang.SortiertEinfügen(dNeu: dNeu)
    }

    /**
    Entfernt einen Knoten mit einem bestimmten Datenelement aus der Liste
    - parameters:
        - dvergleich: gesuchtes Datenelement
    */
    func KnotenEntfernen(dvergleich: Datenelement)
    {
        anfang = anfang.KnotenEntfernen(dvergleich: dvergleich)
    }
}
}

/**
Klasse zur Verwaltung der Lottozahlen

- author: Klaus Reinold
- version: 1.0
*/
class Lotto
{
    private var l: Liste

    /**
	Konstruktor der Klasse Lotto
	erzeugt eine leere Liste
    */
    init()
    {
        l = Liste()
    }
    
    /**
	Einfügen einer neuen Zahl
	- parameters:
		- i: einzufügende int-Zahl
    */
    func Einfügen(i: Zahl)
    {
        let zahl = Zahl(i)
        l.SortiertEinfügen(dNeu: zahl)
    }
    
    
    /**
    Ausgabe aller Zahlen über die Konsole
    */
    func Ausgeben()
    {
        l.InformationAusgeben()
    }
}
